/* $Workfile: OsaLinuxSpi.c $   $Revision: 5 $    $Modtime: 10/30/07 5:31p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc internal

  @module

  @normal Copyright (C) Odin TeleSystems Inc., 2007, All Rights Reserved

*******************************************************************************/

/******************************************************************************
  
  Driver exported entry/exit functions.

  Some of the Linux kernel headers don't like to be compiled with g++ (C++)
  so we need a file that is pure C. This file has two purposes: it implements
  the two Linux module specific entry exit routines, init_module and
  cleanup_module, the file operation callbacks, and also some routines that's
  needed for proc operations.

******************************************************************************/
// This is the only place where this file should be included
// It's unlikely it will be needed anywhere else but if so
// __NO_VERSION__ must be defined before inclusion at the
// other places


// Only include in DaVinci
#ifdef DAVINCI

#include <linux/syscalls.h>
#include <linux/kernel.h>
#include <linux/fcntl.h>
#include <linux/stat.h>
#include <linux/smp_lock.h>


#include <asm/uaccess.h>
#include <asm/system.h>

#include "OsaLinuxSpi.h"

// Path for device
// SPI driver code is in development. The following may
// be defined in the SPI header file for the driver but it is
// not known at this time.
#define OSA_SPI_DEVICE_PATH  "/dev/spi0";
#define OSA_SPI_DEVICE_FLAGS  O_RDWR;

#define IOCTL_SPI_EXCHANGEBUF   18
#define IOCTL_SPI_READBUF       19
#define IOCTL_SPI_WRITEBUF      20


#define OSA_SPI_BUFFER_SIZE   1024 * 16

static int ms_hSpiFileDesc = 0;
static unsigned short ms_nDataBufferCmd[OSA_SPI_BUFFER_SIZE];
static unsigned short ms_nDataBufferStatus[OSA_SPI_BUFFER_SIZE];

#define OSA_SPI_IOCTL spi_dev_ioctl

//
// The SPI driver is still in development. 
// Not sure how to access the ioctl.
// Need a working example.
// Could use the sys_read and sys_write
//
int spi_dev_ioctl(unsigned int a1, unsigned int a2, unsigned long a3) { return -1;}

/******************************************************************************
  Open the SPI device
    return 0=>closed, 1=>open
******************************************************************************/
OsaBool32 OsaLinuxSpiIsOpen(void)
{

  if(ms_hSpiFileDesc) {
    OSA_DPRINTF(("OsaLinuxSpiOpen: SPI is open\n"));
    return OSA_TRUE;
  }
  else {
    OSA_DPRINTF(("OsaLinuxSpiOpen: SPI is closed\n"));
    return OSA_FALSE;
  }
}


/******************************************************************************
  Open the SPI device
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiOpen(void)
{
  int status = 0;
  int  flag = OSA_SPI_DEVICE_FLAGS;
  char path[] = OSA_SPI_DEVICE_PATH;

  if(ms_hSpiFileDesc) {
    OSA_DPRINTF(("OsaLinuxSpiOpen: SPI was already open\n"));
  }
  else {

    ms_hSpiFileDesc = sys_open(path, flag, 0);

    if(!ms_hSpiFileDesc) {
      OSA_DPRINTF(("OsaLinuxSpiOpen: Failed!\n"));
      status = -1;
    }
  }

  OSA_DPRINTF(("OsaLinuxSpiOpen: finished\n"));

  return status;
}

/******************************************************************************
  Close the SPI device. Not sure why this would be needed, OTX driver should
  be active a all times. 
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiClose(void)
{
  int status = 0;

  if(!ms_hSpiFileDesc) {
    OSA_DPRINTF(("OsaLinuxSpiClose: SPI was not open nothing to close\n"));
  }
  else {
    status = sys_close(ms_hSpiFileDesc);
    if(status == -1) {
      OSA_DPRINTF(("OsaLinuxSpiClose: close failed\n"));
    }
  }

  ms_hSpiFileDesc = 0;

  OSA_DPRINTF(("OsaLinuxSpiClose: finished\n"));

  return status;
}


/******************************************************************************
  Read a single word form the SPI
  Note because of SPI access limitation only read local buffer.
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiRead16(OsaUint16 nAddress, OsaUint16 *nData, OsaUint16 nType)
{
  int status = 0;

  // The SPI interface currently only has buffer access.
  // Read the 16K buffer into local memory then access the
  // word of data.

  if(ms_hSpiFileDesc == 0) {
    // SPI is not currently opened.
    OSA_DPRINTF(("OsaLinuxSpiRead16: Error device is not setup\n"));
    status = -1;
  }
  else
  if(nAddress > OSA_SPI_BUFFER_SIZE) {
    OSA_DPRINTF(("OsaLinuxSpiRead16: Address is out of range\n"));
    status = -1;
  }
  else {
#ifndef SPI_BUFFER_ACCESS
    status = OsaLinuxSpiGetReadBuffer(nType);
    if(status == -1) {
      OSA_DPRINTF(("OsaLinuxSpiRead16: SPI read failed\n"));
      *nData = 0;
    }
    else {
      // Copy the data to the user 
      if(nType == OSA_SPI_COMMAND_BUFFER) {
        *nData = ms_nDataBufferCmd[nAddress];
      }
      else {
        *nData = ms_nDataBufferStatus[nAddress];
      }
    }
#else
    // Do not get data from SPI get it from the local buffer. 
    // SPI buffer access is slow.
    if(nType == OSA_SPI_COMMAND_BUFFER) {
      *nData = ms_nDataBufferCmd[nAddress];
    }
    else {
      *nData = ms_nDataBufferStatus[nAddress];
    }

#endif
  }

  return status;
}


/******************************************************************************
  Write a single word to the SPI.
  Note because of SPI access limitation only write to local buffer.
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiWrite16(OsaUint16 nAddress, OsaUint16 nData, OsaUint16 nType)
{
  int status = 0;

  // The SPI interface currently only has buffer access.
  // Read the 16K buffer into local memory then access the
  // word of data.


  if(ms_hSpiFileDesc == 0) {
    // SPI is not currently opened.
    OSA_DPRINTF(("OsaLinuxSpiWrite16: Error device is not setup\n"));
    status = -1;
  }
  else
  if(nAddress > OSA_SPI_BUFFER_SIZE) {
    OSA_DPRINTF(("OsaLinuxSpiWrite16: Address is out of range\n"));
    status = -1;
  }
  else {
#ifndef SPI_BUFFER_ACCESS
    // Get the current data from the SPI buffer and put it 
    // in our local buffer
    status = OsaLinuxSpiGetReadBuffer(nType);
    if(status == -1) {
      OSA_DPRINTF(("OsaLinuxSpiWrite16: SPI read of buffer failed\n"));
    }
    else {
      // Copy the user data to the buffer 
      if(nType == OSA_SPI_COMMAND_BUFFER) {
        ms_nDataBufferCmd[nAddress] = nData;
      }
      else {
        ms_nDataBufferStatus[nAddress] = nData;
      }

      // Write buffer to the SPI
      status = OsaLinuxSpiSendWriteBuffer(nType);
      if(status == -1) {
        OSA_DPRINTF(("OsaLinuxSpiWrite16: SPI write failed\n"));
      }
    }
#else
    // Copy the user data to the buffer 

    // Do not get data from SPI use the local buffer. 
    // SPI buffer access is slow.
    if(nType == OSA_SPI_COMMAND_BUFFER) {
      ms_nDataBufferCmd[nAddress] = nData;
    }
    else {
      ms_nDataBufferStatus[nAddress] = nData;
    }

#endif
  }

  return status;
}


/******************************************************************************
  Read SPI buffer
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiReadBuffer(OsaUint16 *nData, OsaUint16 nType)
{
  unsigned int nIoctlCmd;
  int status = 0;

  if(nType == OSA_SPI_COMMAND_BUFFER) {
    nIoctlCmd = IOCTL_SPI_READBUF;
  }
  else {
    nIoctlCmd = IOCTL_SPI_READBUF;
  }

  if(ms_hSpiFileDesc == 0) {
    // SPI is not currently opened.
    OSA_DPRINTF(("OsaLinuxSpiReadBuffer: Error device is not setup\n"));
    status = -1;
  }
  else {

    status = OSA_SPI_IOCTL(ms_hSpiFileDesc, 
                           nIoctlCmd, 
                           (unsigned long) nData);
    if(status == -1) {
      OSA_DPRINTF(("OsaLinuxSpiReadBuffer: SPI read failed\n"));
    }
  }

  return status;
}


/******************************************************************************
  Write SPI buffer 
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiWriteBuffer(OsaUint16 *nData, OsaUint16 nType)
{
  unsigned int nIoctlCmd;
  int status = 0;

  if(nType == OSA_SPI_COMMAND_BUFFER) {
    nIoctlCmd = IOCTL_SPI_WRITEBUF;
  }
  else {
    nIoctlCmd = IOCTL_SPI_WRITEBUF;
  }

  if(ms_hSpiFileDesc == 0) {
    // SPI is not currently opened.
    OSA_DPRINTF(("OsaLinuxSpiWriteBuffer: Error device is not setup\n"));
    status = -1;
  }
  else {
    // Write buffer to the SPI

    status = OSA_SPI_IOCTL(ms_hSpiFileDesc, 
                           nIoctlCmd, 
                           (unsigned long) nData);
    if(status == -1) {
      OSA_DPRINTF(("OsaLinuxSpiWriteBuffer: SPI write failed\n"));
    }
  }

  return status;
}


/******************************************************************************
  Get the SPI buffer from FPGA shared memory and put in local memory
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiGetReadBuffer(OsaUint16 nType)
{
  unsigned int nIoctlCmd;
  unsigned long nBufferAdd;
  int status = 0;

  if(nType == OSA_SPI_COMMAND_BUFFER) {
    nIoctlCmd = IOCTL_SPI_READBUF;
    nBufferAdd = (unsigned long) ms_nDataBufferCmd;
  }
  else {
    nIoctlCmd = IOCTL_SPI_READBUF;
    nBufferAdd = (unsigned long) ms_nDataBufferStatus;
  }

  if(ms_hSpiFileDesc == 0) {
    // SPI is not currently opened.
    OSA_DPRINTF(("OsaLinuxSpiReadBuffer: Error device is not setup\n"));
    status = -1;
  }
  else {

    status = OSA_SPI_IOCTL(ms_hSpiFileDesc, 
                           nIoctlCmd, 
                           nBufferAdd);
    if(status == -1) {
      OSA_DPRINTF(("OsaLinuxSpiReadBuffer: SPI read failed\n"));
    }
  }

  return status;
}


/******************************************************************************
  Write local SPI buffer to FPGA shared memory
    return 0=>OK, -1=>error
******************************************************************************/
int OsaLinuxSpiSendWriteBuffer(OsaUint16 nType)
{
  unsigned int nIoctlCmd;
  unsigned long nBufferAdd;
  int status = 0;

  asmlinkage long sysStatus;

  if(nType == OSA_SPI_COMMAND_BUFFER) {
    nIoctlCmd = IOCTL_SPI_WRITEBUF;
    nBufferAdd = (unsigned long) ms_nDataBufferCmd;
  }
  else {
    nIoctlCmd = IOCTL_SPI_WRITEBUF;
    nBufferAdd = (unsigned long) ms_nDataBufferStatus;
  }

  if(ms_hSpiFileDesc == 0) {
    // SPI is not currently opened.
    OSA_DPRINTF(("OsaLinuxSpiWriteBuffer: Error device is not setup\n"));
    status = -1;
  }
  else {
    // Write buffer to the SPI

    sysStatus = OSA_SPI_IOCTL((unsigned int) ms_hSpiFileDesc, 
                              (unsigned int) nIoctlCmd, 
                              (unsigned long) nBufferAdd);

    if(sysStatus == -1) {
      OSA_DPRINTF(("OsaLinuxSpiWriteBuffer: SPI write failed\n"));
      status = -1;
    }
  }

  return status;
}




#endif // DAVINCI
